#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
try:
    from ._version import get_version

    VERSION = get_version()

except ImportError:
    from ._version import get_versions

    VERSION = get_versions()["version"]


RELEASE_TAG = ""


def safe_int(x):
    try:
        return int(x)
    except ValueError:
        return x


# I won't put the release tag in the version_info tuple.  Since PEP440 is on
# the way.
VERSION_INFO = tuple(safe_int(x) for x in VERSION.split("."))


del safe_int



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

# TODO: Document (version.rst) this module.  Add tests.
'''Versions API

'''

from xotl.tools.decorator import singleton


def _check(info):
    '''Validate a version info.

    :param info: could be a string, an integer, float, or any integer
           collection (only first three valid integers are used).

    :returns: a valid tuple or an error if invalid.

    '''
    from collections import Iterable
    from distutils.version import LooseVersion, StrictVersion
    MAX_COUNT = 3
    if isinstance(info, (int, float)):
        aux = str(info)
    elif isinstance(info, Iterable) and not isinstance(info, str):
        aux = '.'.join(map(str, info))
    else:
        aux = info
    if isinstance(aux, str):
        try:
            essay = StrictVersion(aux)
        except (TypeError, ValueError):    # Being as safe as possible.
            essay = LooseVersion(aux)
        res = essay.version[:MAX_COUNT]
        if any(res):
            return tuple(res)
        else:
            raise ValueError("invalid version value '{}'".format(info))
    else:
        msg = "Invalid type '{}' for version '{}'"
        raise TypeError(msg.format(type(info).__name__, info))


class ThreeNumbersVersion(tuple):
    '''Structured version info considering valid first 3 members

    This class is mainly intended to be sub-classed as a singleton resulting
    in a tuple with three integer components as 'major', 'minor', and 'micro'.

    Instances of this class can be compared with a variety of value types:

    - An integer with the 'major' component.

    - A float with the ('major', 'minor') components.

    - A string is converted to a version tuple before compare it.

    - Any collection with a prefix of at least three logical integers (that is
      ``[1.3, '2a']`` is the same as ``'1.3.2a'`` and ``(1, 3, 2)``).

    Equality comparison is relevant only for heading values: ``3.1 == 3``.

    .. versionadded:: 1.8.0

    '''
    def __new__(cls, info):
        MAX_COUNT = 3
        head = _check(info)
        tail = (0,) * (MAX_COUNT - len(head))
        return super().__new__(cls, head + tail)

    @property
    def major(self):
        return self[0]

    @property
    def minor(self):
        return self[1]

    @property
    def micro(self):
        return self[2]

    def to_float(self):
        return float('{}.{}'.format(*self[:2]))

    __float__ = to_float
    __trunc__ = major

    def __eq__(self, other):
        try:
            aux = _check(other)
            this = self[:len(aux)]
            return this == aux
        except (TypeError, ValueError):
            return False

    def __lt__(self, other):
        try:
            return tuple(self) < _check(other)
        except (TypeError, ValueError):
            return NotImplemented

    def __gt__(self, other):
        try:
            return tuple(self) > _check(other)
        except (TypeError, ValueError):
            return NotImplemented

    def __ne__(self, other):
        return not (self == other)

    def __le__(self, other):
        return not (self > other)

    def __ge__(self, other):
        return not (self < other)


@singleton
class python_version(ThreeNumbersVersion):
    '''Current Python version.

    Initialized with `~sys.version_info`:obj: 5 components tuple.

    Extra components (besides 'major', 'minor', and 'micro') are:
    'releaselevel' (a string that could be 'alpha', 'beta', 'candidate', or
    'final'), and 'serial' (an integer).  The attribute 'pypy' could be used
    to determine if this is a PyPy instance or not.

    '''
    def __new__(cls):
        import sys
        self = super().__new__(cls, sys.version_info)
        self.pypy = sys.version.find('PyPy') >= 0
        return self

    @property
    def releaselevel(self):
        return self[3]

    @property
    def serial(self):
        return self[4]


def _get_mod_version(mod):
    '''Get a valid version from a module.

    Used internally by `PackageVersion`:class:.

    '''
    valid_names = ('VERSION_INFO', 'VERSION', 'version_info', 'version',
                   '__VERSION__', '__version__')
    i = 0
    res = None
    while res is None and i < len(valid_names):
        name = valid_names[i]
        version = getattr(mod, name, None)
        if version is not None:
            try:
                res = _check(version)
            except (TypeError, ValueError):
                pass
        i += 1
    if not res:
        import os
        path = os.path.dirname(os.__file__)
        if mod.__file__.startswith(path):
            res = python_version
    return res


class PackageVersion(ThreeNumbersVersion):
    '''Current Package version.

    Extra components (besides 'major', 'minor', and 'micro') are:
    'releaselevel' (a string that could be 'alpha', 'beta', 'candidate', or
    'final'), and 'serial' (an integer).  The attribute 'pypy' could be used
    to determine if this is a PyPy instance or not.

    '''
    def __new__(cls, package_name):
        info = cls._find_version(package_name)
        if info:
            return super().__new__(cls, info)
        else:
            msg = '{}() could not determine a valid version'
            raise ValueError(msg.format(cls.__name__))

    @staticmethod
    def _find_version(package_name):
        from pkg_resources import get_distribution, ResolutionError
        if package_name in ('__builtin__', 'builtins'):
            return python_version
        else:
            res = None
            while not res and package_name:
                try:
                    dist = get_distribution(package_name)
                    try:
                        res = dist.parsed_version.base_version
                    except AttributeError:
                        res = dist.version
                except ResolutionError:
                    from importlib import import_module
                    try:
                        mod = import_module('.'.join(
                            (package_name, 'release')
                        ))
                        res = _get_mod_version(mod)
                    except ImportError:
                        try:
                            mod = import_module(package_name)
                            res = _get_mod_version(mod)
                        except ImportError:
                            mod = __import__(package_name)
                            res = _get_mod_version(mod)
                if not res:
                    aux = package_name.rsplit('.', 1)
                    package_name = aux[0] if len(aux) > 1 else ''
            return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Synchronization lock tools.

A lock object is a synchronization primitive.  Locks are used as Python
Context Managers with only one method (`enter`) and one property (`locked`).

The method `enter` is used with the Python ``with`` statement and the property
`locked` is logically True or False depending if the lock is active or not.

For example::

  >>> from xotl.tools.lock import context_lock as ctx

  >>> def get_lock_value():
  ...     return 'Yes' if ctx.locked else 'No'

  >>> with ctx.enter():
  ...     one = get_lock_value()
  >>> two = get_lock_value()
  >>> (one, two)
  ('Yes', 'No')


Locks are implemented using module-property; this means that each time you
import it, a different lock is returned::

  >>> from xotl.tools.tasking.lock import context_lock as one
  >>> from xotl.tools.tasking.lock import context_lock as two
  >>> one is two
  False

The function `context_lock`:func: implement a module property to create a
class that use an execution context, see `xotl.tools.context`:mod: module for
more information.

If other lock mechanisms are going to be implementing, for example using
threading, this is the place.

'''

from xotl.tools.modules import moduleproperty


@moduleproperty
def context_lock(self):
    '''Allocate a lock based on xoutil execution contexts.'''
    from xotl.tools.objects import classproperty

    class ContextLock:
        '''A class representing the lock.

        See `xotl.tools.lock`:mod: module for more information.

        '''
        def __new__(cls, *args, **kwargs):
            msg = '"{}" could not be instanced.'.format(cls.__name__)
            raise RuntimeError(msg)

        @classmethod
        def enter(cls, **kwargs):
            '''Enter the context.'''
            from xotl.tools.context import context
            return context(cls, **kwargs)

        @classproperty
        def locked(cls):
            from xotl.tools.context import context
            return context[cls]

    return ContextLock


del moduleproperty



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Extend standard modules including "future" features in current versions.

Version 3 introduce several concepts in standard modules.  Sometimes these
features are implemented in the evolution of 2.7.x versions.  By using
sub-modules, these differences can be avoided transparently.  For example, you
can import `xotl.tools.future.collections.UserDict`:class: in any version,
that it's equivalent to Python 3 `collections.UserDict`:class:, but it don't
exists in Python 2.

.. versionadded:: 1.7.2

'''



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Extensions to Python's `time` module.

You may use it as drop-in replacement of `time`.  Although we don't
document all items here.  Refer to `time`:mod: documentation.

.. note:: This module is deprecated since `monotonic` is included in
Python 3.3.

'''

from time import *  # noqa
from time import monotonic
import time as _stdlib    # noqa

from xotl.tools.deprecation import deprecate_module
deprecate_module(replacement=monotonic.__module__)
del deprecate_module



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''
Regular expressions and validation functions for several identifiers.
'''

from re import compile as _regex_compile


__all__ = ('is_valid_identifier', 'is_valid_full_identifier',
           'is_valid_public_identifier',
           'is_valid_slug')


def is_valid_identifier(name):
    '''Returns True if `name` a valid Python identifier.

    If `name` is not a string, return False.  This is roughly::

        isinstance(name, str) and name.isidentifier()

    '''
    return isinstance(name, str) and name.isidentifier()


def check_identifier(name):
    '''Checks if `name` a valid Python identifier.

    If not, an exception is raised.

    '''
    if is_valid_identifier(name):
        return name
    else:
        raise ValueError('"%s" is not a valid identifier!' % name)


def is_valid_full_identifier(name):
    '''Returns True if `name` is a valid dotted Python identifier.

    See `is_valid_identifier`:func: for what "validity" means.

    '''
    if isinstance(name, str):
        return all(part.isidentifier() for part in name.split('.'))
    else:
        return False


def is_valid_public_identifier(name):
    '''Returns True if `name` is a valid Python identifier that is deemed
    public.

    Convention says that any name starting with a "_" is not public.

    See `is_valid_identifier`:func: for what "validity" means.

    '''
    return is_valid_identifier(name) and not name.startswith('_')


_SLUG_REGEX = _regex_compile(r'(?i)^[\w]+([-][\w]+)*$')


def is_valid_slug(slug):
    return isinstance(slug, str) and _SLUG_REGEX.match(slug)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Tool to show a progress percent in the terminal.

.. deprecated:: 2.1.0

'''


__all__ = ['Progress']


_HELIX = '|/-\\'


class Progress:
    '''
    Print a progress percent to the console. Also the elapsed and the
    estimated times.

    To signal an increment in progress just call the instance and (optionally)
    pass a message like in::

        progress = Progress(10)
        for i in range(10):
            progress()
    '''

    def __init__(self, max_value=100, delta=1, first_message=None,
                 display_width=None):
        from xotl.tools.future.datetime import datetime
        self.max_value = max_value
        self.delta = delta
        self.percent = self.value = 0
        self.start_time = datetime.now()
        self.first_message = first_message
        self.display_width = display_width

    def __call__(self, progress=None, message='...'):
        if self.first_message is not None:
            print(self.first_message)
            self.first_message = None
        if progress is None:
            self.value += self.delta
        else:
            self.value = progress
        percent = 100 * self.value // self.max_value
        if self.percent != percent:
            import sys
            from xotl.tools.future.datetime import strfdelta
            self.percent = percent
            helix = _HELIX[percent % len(_HELIX)]
            elapsed = self.start_time.now() - self.start_time
            _cls = type(elapsed)
            total = _cls(seconds=elapsed.total_seconds() * 100 / self.percent)
            _fmt = '\r{helix} {percent}% - "{elapsed}" of about "{total}"'
            progress_line = _fmt.format(helix=helix, percent=percent,
                                        elapsed=strfdelta(elapsed),
                                        total=strfdelta(total))
            max_width = self.display_width or self._get_terminal_width()
            _fmt = ('{message: >%d}' % (max_width - len(progress_line) - 1))
            progress_line += _fmt.format(message=message)
            print(progress_line, end=('' if percent != 100 else '\n\r'))
            sys.stdout.flush()

    def _get_terminal_width(self, default=120):
        import os
        try:
            return int(os.environ.get('COLUMNS', default))
        except ValueError:
            return default
